"""Methods for filesystem-like operations."""

from __future__ import annotations

import datetime
import json
from pathlib import Path
from typing import TYPE_CHECKING, Literal, overload

import httpx

from ._remote_methods import ArtifactMethod
from ._utils import (
    check_errors,
    filter_by_name,
    get_headers,
    get_method_url,
    params_list_files,
    params_remove_file,
    prepare_params,
    walk_dir,
)

if TYPE_CHECKING:
    from . import AsyncHyphaArtifact
if TYPE_CHECKING:
    from hypha_artifact.classes import ArtifactItem


@overload
async def ls(
    self: AsyncHyphaArtifact,
    path: str = ".",
    version: str | None = None,
    *,
    detail: Literal[False],
) -> list[str]: ...


@overload
async def ls(
    self: AsyncHyphaArtifact,
    path: str = ".",
    version: str | None = None,
    *,
    detail: Literal[True],
) -> list[ArtifactItem]: ...


@overload
async def ls(
    self: AsyncHyphaArtifact,
    path: str = ".",
    version: str | None = None,
    *,
    detail: None | bool = True,
) -> list[ArtifactItem]: ...


async def ls(
    self: AsyncHyphaArtifact,
    path: str = ".",
    version: str | None = None,
    *,
    detail: None | bool = False,
) -> list[str] | list[ArtifactItem]:
    """List contents of path.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The HyphaArtifact instance to use
    path: str
        Path to list contents of
    detail: bool | None
        Whether to include detailed information about each item
    version: str | None
        The version of the artifact to list contents from.
        By default, it lists from the latest version.
        If you want to list from a staged version, you can set it to "stage".

    Returns
    -------
    list[str] | list[ArtifactItem]
        List of file names or detailed artifact items

    """
    simple_params = params_list_files(dir_path=path, version=version)
    params = prepare_params(self, simple_params)

    url = get_method_url(self, ArtifactMethod.LIST_FILES)

    response = await self.get_client().get(
        url,
        params=params,
        headers=get_headers(self),
        timeout=60,
    )

    check_errors(response)

    artifact_items: list[ArtifactItem] = json.loads(response.content)

    if detail:
        return artifact_items

    return [item["name"] for item in artifact_items]


async def info(
    self: AsyncHyphaArtifact,
    path: str,
    version: str | None = None,
) -> ArtifactItem:
    """Get information about a file or directory.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The HyphaArtifact instance to use
    path: str
        Path to get information about
    version:
        The version of the artifact to get the information from.
        By default, it reads from the latest version.
        If you want to read from a staged version, you can set it to "stage".

    Returns
    -------
    dict
        Dictionary with file information

    """
    parent_path = str(Path(path).parent)
    files_here = await self.ls(parent_path, detail=True, version=version)
    matching_files_here = filter_by_name(files_here, path)

    if matching_files_here:
        return matching_files_here[0]

    files_in_sub = await self.ls(path, detail=True, version=version)
    matching_files_in_sub = filter_by_name(files_in_sub, path)

    if len(matching_files_in_sub) == 1:
        return matching_files_in_sub[0]

    if len(matching_files_in_sub) > 1 or files_in_sub:
        return {"name": path, "type": "directory", "size": 0, "last_modified": None}

    raise FileNotFoundError(path)


async def isdir(
    self: AsyncHyphaArtifact,
    path: str,
    version: str | None = None,
) -> bool:
    """Check if a path is a directory.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The HyphaArtifact instance to use
    path: str
        Path to check
    version: str | None = None
        The version of the artifact to check against.
        By default, it checks the latest version.
        If you want to check a staged version, you can set it to "stage".

    Returns
    -------
    bool
        True if the path is a directory, False otherwise

    """
    try:
        path_info = await self.info(path, version=version)
        return path_info["type"] == "directory"
    except (OSError, FileNotFoundError):
        return False


async def isfile(
    self: AsyncHyphaArtifact,
    path: str,
    version: str | None = None,
) -> bool:
    """Check if a path is a file.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The HyphaArtifact instance to use
    path: str
        Path to check
    version: str | None = None
        The version of the artifact to check against.
        By default, it checks the latest version.
        If you want to check a staged version, you can set it to "stage".

    Returns
    -------
    bool
        True if the path is a file, False otherwise

    """
    try:
        path_info = await self.info(path, version=version)
        return path_info["type"] == "file"
    except (OSError, FileNotFoundError):
        return False


async def listdir(
    self: AsyncHyphaArtifact,
    path: str,
    version: str | None = None,
) -> list[str]:
    """List files in a directory.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The HyphaArtifact instance to use
    path: str
        Path to list
    version: str | None = None
        The version of the artifact to get the information from.
        By default, it reads from the latest version.
        If you want to read from a staged version, you can set it to "stage".

    Returns
    -------
    list of str
        List of file names in the directory

    """
    return await self.ls(path, detail=False, version=version)


@overload
async def find(
    self: AsyncHyphaArtifact,
    path: str,
    maxdepth: int | None = None,
    version: str | None = None,
    *,
    withdirs: bool = False,
    detail: Literal[True],
    hide_keep: bool = True,
) -> dict[str, ArtifactItem]: ...


@overload
async def find(
    self: AsyncHyphaArtifact,
    path: str,
    maxdepth: int | None = None,
    version: str | None = None,
    *,
    withdirs: bool = False,
    detail: Literal[False] = False,
    hide_keep: bool = True,
) -> list[str]: ...


async def find(
    self: AsyncHyphaArtifact,
    path: str,
    maxdepth: int | None = None,
    version: str | None = None,
    *,
    withdirs: bool = False,
    detail: bool = False,
    hide_keep: bool = True,
) -> list[str] | dict[str, ArtifactItem]:
    """Find all files (and optional directories) under a path.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The AsyncHyphaArtifact instance to use
    path: str
        Base path to search from
    maxdepth: int or None
        Maximum recursion depth when searching
    withdirs: bool
        Whether to include directories in the results
    detail: bool
        If True, return a dict of {path: info_dict}
        If False, return a list of paths
    version: str | None
        The version of the artifact to search in.
        By default, it searches in the latest version.
        If you want to search in a staged version, you can set it to "stage".
    hide_keep: bool
        If True, exclude .keep files from the results.
        If False, include .keep files in the results.

    Returns
    -------
    list or dict
        List of paths or dict of {path: info_dict}

    """
    filtered_all_files = await walk_dir(
        self,
        path,
        maxdepth,
        1,
        version,
        withdirs=withdirs,
    )

    filtered_all_files = (
        {k: v for k, v in filtered_all_files.items() if not k.endswith(".keep")}
        if hide_keep
        else filtered_all_files
    )

    if detail:
        return filtered_all_files

    return sorted(filtered_all_files.keys())


async def modified(
    self: AsyncHyphaArtifact,
    path: str,
    version: str | None = None,
) -> datetime.datetime | None:
    """Return the modified timestamp of a file as a datetime.datetime.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The AsyncHyphaArtifact instance to use
    path: str
        Path to the file
    version: str | None = None
        The version of the artifact to check against.
        By default, it checks the latest version.
        If you want to check a staged version, you can set it to "stage".

    Returns
    -------
    datetime or None
        Modified time of the file, if available

    """
    path_info = await self.info(path, version=version)

    last_modified = path_info["last_modified"]

    if last_modified:
        return datetime.datetime.fromtimestamp(
            last_modified,
            tz=datetime.UTC,
        )

    return None


async def size(
    self: AsyncHyphaArtifact,
    path: str,
    version: str | None = None,
) -> int:
    """Get the size of a file in bytes.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The AsyncHyphaArtifact instance to use
    path: str
        Path to the file
    version: str | None = None
        The version of the artifact to check against.
        By default, it checks the latest version.
        If you want to check a staged version, you can set it to "stage".

    Returns
    -------
    int
        Size of the file in bytes

    """
    path_info = await self.info(path, version=version)
    if path_info["type"] == "directory":
        return 0
    return int(path_info["size"])


async def sizes(
    self: AsyncHyphaArtifact,
    paths: list[str],
    version: str | None = None,
) -> list[int]:
    """Get the size of multiple files.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The AsyncHyphaArtifact instance to use
    paths: list of str
        List of paths to get sizes for
    version: str | None = None
        The version of the artifact to check against.
        By default, it checks the latest version.
        If you want to check a staged version, you can set it to "stage".

    Returns
    -------
    list of int
        List of file sizes in bytes

    """
    return [await self.size(path, version=version) for path in paths]


async def rm(
    self: AsyncHyphaArtifact,
    path: str,
    maxdepth: int | None = None,
    *,
    recursive: bool = False,
) -> None:
    """Remove file or directory.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The AsyncHyphaArtifact instance to use
    path: str
        Path to the file or directory to remove
    recursive: bool
        Defaults to False. If True and path is a directory,
        remove all its contents recursively
    maxdepth: int or None
        Maximum recursion depth when recursive=True

    Returns
    -------
    datetime or None
        Creation time of the file, if available

    """
    paths_to_remove: list[str] = []
    is_dir = await self.isdir(path)
    if recursive and is_dir:
        paths_to_remove = await self.find(
            path,
            maxdepth=maxdepth,
            withdirs=False,
            detail=False,
            hide_keep=False,
        )
    elif not recursive and is_dir:
        error_msg = (
            f"Path is a directory: {path}. Use --recursive to remove directories."
        )
        raise IsADirectoryError(error_msg)
    else:
        paths_to_remove.append(path)

    for file_path in paths_to_remove:
        simple_params = params_remove_file(file_path)
        params = prepare_params(self, simple_params)
        response = await self.get_client().post(
            url=get_method_url(self, ArtifactMethod.REMOVE_FILE),
            headers=get_headers(self),
            json=params,
        )

        check_errors(response)


async def rm_file(self: AsyncHyphaArtifact, path: str) -> None:
    """Remove a file.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The AsyncHyphaArtifact instance to use
    path: str
        Path to remove

    """
    await self.rm(path)


async def rmdir(self: AsyncHyphaArtifact, path: str) -> None:
    """Remove an empty directory.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The AsyncHyphaArtifact instance to use
    path: str
        Path to remove

    """
    if not await self.isdir(path):
        error_msg = f"Directory not found: {path}"
        raise FileNotFoundError(error_msg)

    file_names = await self.ls(path, detail=False)
    has_keep = any(f_name == ".keep" for f_name in file_names)
    if (not has_keep and len(file_names) > 0) or (has_keep and len(file_names) > 1):
        error_msg = f"Directory not empty: {path}"
        raise OSError(error_msg)

    await self.rm(str(Path(path) / ".keep"))


async def touch(
    self: AsyncHyphaArtifact,
    path: str,
    *,
    truncate: bool = True,
) -> None:
    """Create a file if it does not exist, or update its last modified time.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The AsyncHyphaArtifact instance to use
    path: str
        Path to the file
    truncate: bool
        If True, always set file size to 0;
        if False, update timestamp and leave file unchanged

    """
    async with self.open(path, "wb") as f:
        if truncate or not await self.exists(path):
            return

        if not truncate:
            current_content = await f.read()
            f.seek(0)
            await f.write(current_content)


async def mkdir(
    self: AsyncHyphaArtifact,
    path: str,
    *,
    create_parents: bool = True,
) -> None:
    """Create a directory.

    Creates a .keep file in the directory to ensure it exists.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The AsyncHyphaArtifact instance to use
    path: str
        Path to create
    create_parents: bool
        If True, create parent directories if they don't exist

    """
    if Path(path) == Path():
        return

    parent_path = str(Path(path).parent)
    child_path = str(Path(path).name)

    if parent_path and not await self.exists(parent_path):
        if not create_parents:
            error_msg = f"Parent directory does not exist: {parent_path}"
            raise FileNotFoundError(error_msg)

        await self.mkdir(parent_path, create_parents=True)

    if parent_path and await self.isfile(parent_path):
        error_msg = f"Parent path is not a directory: {parent_path}"
        raise NotADirectoryError(error_msg)

    await self.touch(str(Path(child_path) / ".keep"))


async def makedirs(
    self: AsyncHyphaArtifact,
    path: str,
    *,
    exist_ok: bool = True,
) -> None:
    """Recursively make directories.

    Creates directory at path and any intervening required directories.
    Raises exception if, for instance, the path already exists but is a
    file.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The HyphaArtifact instance to use
    path: str
        Path to create
    exist_ok: bool
        If False and the directory exists, raise an error

    """
    if not exist_ok and await self.exists(path):
        error_msg = f"Directory already exists: {path}"
        raise FileExistsError(error_msg)

    await self.mkdir(path, create_parents=True)


async def exists(
    self: AsyncHyphaArtifact,
    path: str,
    version: str | None = None,
) -> bool:
    """Check if a file or directory exists.

    Parameters
    ----------
    self: AsyncHyphaArtifact
        The HyphaArtifact instance to use
    path: str
        Path to check
    version: str | None
        The version of the artifact to check against. If None, uses the latest version.

    Returns
    -------
    bool
        True if the path exists, False otherwise

    """
    try:
        async with self.open(path, "r", version=version) as f:
            await f.read(0)
            return True
    except (OSError, FileNotFoundError, httpx.HTTPStatusError, httpx.RequestError):
        try:
            dir_files = await self.ls(path, detail=False, version=version)
            return len(dir_files) > 0
        except (OSError, FileNotFoundError, httpx.HTTPStatusError, httpx.RequestError):
            return False



"""Utility functions for async hypha artifact."""

from __future__ import annotations

import asyncio
import math
import os
from http import HTTPStatus
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    Protocol,
    Self,
    TypedDict,
    cast,
    runtime_checkable,
)

import httpx

from hypha_artifact.async_hypha_artifact._remote_methods import ArtifactMethod
from hypha_artifact.classes import MultipartStatusMessage

if TYPE_CHECKING:
    from collections.abc import Callable, Mapping
    from types import TracebackType

    from hypha_artifact.classes import ArtifactItem

    from . import AsyncHyphaArtifact

MAXIMUM_MULTIPART_THRESHOLD = 100 * 1024 * 1024  # 100 MB
DEFAULT_CHUNK_SIZE = 6 * 1024 * 1024  # 6 MB
MINIMUM_CHUNK_SIZE = 5 * 1024 * 1024  # 5 MB

anyio: Any | None
try:  # pragma: no cover - import guard
    import anyio as _anyio

    anyio = _anyio
    _has_anyio = True
except ImportError:  # pragma: no cover - optional dependency
    anyio = None
    _has_anyio = False


@runtime_checkable
class AsyncBinaryFile(Protocol):
    async def __aenter__(self) -> Self: ...

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None: ...

    async def read(self) -> bytes: ...
    async def write(self, data: bytes) -> int: ...
    async def close(self) -> None: ...


class AsyncFile:
    """Minimal async wrapper around a file object.

    Provides async context management and async read/write using a thread
    offload when anyio is unavailable. When anyio is present, use anyio's
    async file operations directly.
    """

    def __init__(self, fp: object, mode: str) -> None:
        self._fp = fp
        self._mode = mode

    async def __aenter__(self) -> Self:
        return self

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        await self.close()

    async def read(self) -> bytes:
        f = cast("Any", self._fp)
        return await asyncio.to_thread(f.read)

    async def write(self, data: bytes) -> int:
        f = cast("Any", self._fp)
        return await asyncio.to_thread(f.write, data)

    async def close(self) -> None:
        f = cast("Any", self._fp)
        return await asyncio.to_thread(f.close)


async def aio_open(path: str | Path, mode: str) -> AsyncBinaryFile:
    """Open a file for async usage.

    - If anyio is installed, use anyio.open_file which provides true async I/O
      on supported platforms.
    - Otherwise, open the file synchronously and wrap it so read/write happen
      via asyncio.to_thread to avoid blocking the event loop.
    """
    if _has_anyio and anyio is not None:
        # anyio.open_file returns an async file object implementing aclose/read/write
        open_file_fn = anyio.open_file
        return await open_file_fn(str(path), mode)
    # Fallback: open in a thread; return an AsyncFile wrapper
    fp = await asyncio.to_thread(Path(path).open, mode)
    return AsyncFile(fp, mode)


class SyncBinaryFile(Protocol):
    def read(self, size: int = ...) -> bytes: ...
    def write(self, data: bytes) -> int: ...
    def close(self) -> None: ...


class ListFilesParams(TypedDict, total=False):
    dir_path: str
    version: str


class GetFileUrlParams(TypedDict, total=False):
    file_path: str
    version: str
    use_proxy: bool
    use_local_url: bool | str


class RemoveFileParams(TypedDict, total=False):
    file_path: str


class UploadPartServerInfo(TypedDict):
    """Server-provided info for a part to upload."""

    url: str
    part_number: int


class PreparedPartInfo(TypedDict):
    """Client-prepared part info with data to upload."""

    url: str
    part_number: int
    chunk: bytes
    part_size: int


class CompletedPart(TypedDict):
    """Completed part info used to finalize multipart upload."""

    part_number: int
    etag: str


def params_list_files(
    dir_path: str = ".",
    version: str | None = None,
) -> ListFilesParams:
    """Typed builder for List Files parameters."""
    p: ListFilesParams = {"dir_path": dir_path}
    if version is not None:
        p["version"] = version
    return p


def params_get_file_url(
    file_path: str,
    *,
    version: str | None = None,
    use_proxy: bool | None = None,
    use_local_url: bool | str | None = None,
) -> GetFileUrlParams:
    """Typed builder for GET/PUT file URL params used by fsspec_open and uploads."""
    p: GetFileUrlParams = {"file_path": file_path}
    if version is not None:
        p["version"] = version
    if use_proxy is not None:
        p["use_proxy"] = use_proxy
    if use_local_url is not None:
        p["use_local_url"] = use_local_url
    return p


def params_remove_file(file_path: str) -> RemoveFileParams:
    return {"file_path": file_path}


def params_create(
    *,
    alias: str,
    workspace: str | None,
    parent_id: str | None,
    artifact_type: str | None,
    manifest: str | dict[str, Any] | None,
    config: dict[str, Any] | None,
    version: str | None,
    stage: bool | None,
    comment: str | None,
    secrets: dict[str, str] | None,
    overwrite: bool | None,
) -> dict[str, Any]:
    tmp: dict[str, Any | None] = {
        "alias": alias,
        "workspace": workspace,
        "parent_id": parent_id,
        "type": artifact_type,
        "manifest": manifest,
        "config": config,
        "version": version,
        "stage": stage,
        "comment": comment,
        "secrets": secrets,
        "overwrite": overwrite,
    }
    result: dict[str, Any] = {k: v for k, v in tmp.items() if v is not None}
    return result


def params_put_file_start_multipart(
    file_path: str,
    *,
    part_count: int,
    download_weight: float = 1.0,
    use_proxy: bool | None = None,
    use_local_url: bool | str | None = None,
) -> dict[str, Any]:
    p: dict[str, Any] = {
        "file_path": file_path,
        "part_count": part_count,
        "download_weight": download_weight,
    }
    if use_proxy is not None:
        p["use_proxy"] = use_proxy
    if use_local_url is not None:
        p["use_local_url"] = use_local_url
    return p


def params_put_file_complete_multipart(
    upload_id: str,
    *,
    parts: list[CompletedPart],
) -> dict[str, Any]:
    return {"upload_id": upload_id, "parts": parts}


def params_edit(
    *,
    manifest: dict[str, Any] | None = None,
    type: str | None = None,  # noqa: A002
    config: dict[str, Any] | None = None,
    secrets: dict[str, str] | None = None,
    version: str | None = None,
    comment: str | None = None,
    stage: bool = False,
) -> dict[str, Any]:
    res: dict[str, Any] = {}
    if manifest is not None:
        res["manifest"] = manifest
    if type is not None:
        res["type"] = type
    if config is not None:
        res["config"] = config
    if secrets is not None:
        res["secrets"] = secrets
    if version is not None:
        res["version"] = version
    if comment is not None:
        res["comment"] = comment
    if stage:
        res["stage"] = stage
    return res


def params_commit(
    *,
    version: str | None = None,
    comment: str | None = None,
) -> dict[str, Any]:
    res: dict[str, Any] = {}
    if version is not None:
        res["version"] = version
    if comment is not None:
        res["comment"] = comment
    return res


def params_delete(
    *,
    delete_files: bool | None = None,
    recursive: bool | None = None,
    version: str | None = None,
) -> dict[str, Any]:
    res: dict[str, Any] = {}
    if delete_files is not None:
        res["delete_files"] = delete_files
    if recursive is not None:
        res["recursive"] = recursive
    if version is not None:
        res["version"] = version
    return res


def local_file_or_dir(src_path: str, dst_path: str) -> str:
    """Resolve destination semantics without using the local filesystem.

    - If `dst_path` ends with a path separator (`/` on POSIX), treat it as a
        directory hint and append the basename of `src_path`.
    - Otherwise, treat `dst_path` as the full target path.

    This avoids surprising behavior when a local directory happens to share
    the same name as the intended file (e.g., a directory named 'file.txt').
    """
    is_dir_hint = str(dst_path).endswith(("/", os.sep))
    return str(Path(dst_path) / Path(src_path).name) if is_dir_hint else str(dst_path)


async def remote_file_or_dir(
    self: AsyncHyphaArtifact,
    src_path: str,
    dst_path: str,
) -> str:
    """Resolve remote destination semantics with explicit hint or remote check.

    - If `dst_path` ends with a path separator, treat as directory and append basename.
    - Else, if the remote `dst_path` currently exists as a directory, append basename.
    - Otherwise, treat `dst_path` as the full target path.
    """
    if str(dst_path).endswith(("/", os.sep)):
        return str(Path(dst_path) / Path(src_path).name)
    is_remote_dir = await self.isdir(dst_path)
    return str(Path(dst_path) / Path(src_path).name) if is_remote_dir else str(dst_path)


def env_override(
    env_var_name: str,
    *,
    override: bool | str | None = None,
) -> bool | str | None:
    env_var_val = os.getenv(env_var_name)

    if override is not None:
        return override

    if env_var_val is not None:
        if env_var_val.lower() == "true":
            return True
        return env_var_val

    return None


def to_bytes(content: str | bytes | bytearray | memoryview) -> bytes:
    if isinstance(content, bytes):
        return content
    if isinstance(content, str):
        return content.encode("utf-8")
    return bytes(content)


def decode_to_text(content: str | bytes | bytearray | memoryview) -> str:
    if isinstance(content, str):
        return content
    if isinstance(content, bytes):
        return content.decode("utf-8")
    return bytes(content).decode("utf-8")


def filter_by_name(
    files: list[ArtifactItem],
    name: str,
) -> list[ArtifactItem]:
    """Filter files by name."""
    return [f for f in files if Path(f["name"]).name == Path(name).name]


async def download_to_path(
    self: AsyncHyphaArtifact,
    remote_path: str,
    local_path: str,
    *,
    version: str | None = None,
) -> None:
    parent = Path(local_path).parent
    if parent:
        parent.mkdir(parents=True, exist_ok=True)
    async with self.open(remote_path, "rb", version=version) as src_file:
        data = await src_file.read()
    pre_dst_file = await aio_open(local_path, "wb")
    async with pre_dst_file as dst_file:
        await dst_file.write(to_bytes(data))


async def upload_file_simple(
    self: AsyncHyphaArtifact,
    local_path: str | Path,
    remote_path: str,
) -> None:
    pre_src_file = await aio_open(local_path, "rb")
    async with pre_src_file as src_file:
        data = await src_file.read()
    async with self.open(remote_path, "wb") as dst_file:
        await dst_file.write(data)


async def build_remote_to_local_pairs(
    self: AsyncHyphaArtifact,
    rpath: str | list[str],
    lpath: str | list[str] | None,
    *,
    recursive: bool,
    maxdepth: int | None,
    version: str | None,
) -> list[tuple[str, str]]:
    """Expand rpath/lpath into concrete (remote, local) file pairs.

    Applies recursive listing when asked and errors when a directory is passed
    without recursive flag.
    """
    if not lpath:
        lpath = rpath
    rpaths, lpaths = ensure_equal_len(rpath, lpath)
    pairs: list[tuple[str, str]] = []
    for rp, lp in zip(rpaths, lpaths, strict=False):
        if await self.isdir(rp, version=version):
            if not recursive:
                msg = f"Path is a directory: {rp}. Use --recursive to get directories."
                raise IsADirectoryError(msg)
            Path(lp).mkdir(parents=True, exist_ok=True)
            files = await self.find(
                rp,
                maxdepth=maxdepth,
                withdirs=False,
                version=version,
            )
            pairs.extend(rel_path_pairs(files, src_path=rp, dst_path=lp))
        else:
            pairs.append((rp, lp))
    return pairs


def build_local_to_remote_pairs(
    lpath: str | list[str],
    rpath: str | list[str] | None,
    *,
    recursive: bool,
    maxdepth: int | None,
) -> list[tuple[str, str]]:
    """Expand lpath/rpath into concrete (local, remote) file pairs."""
    if not rpath:
        rpath = lpath
    rpaths, lpaths = ensure_equal_len(rpath, lpath)
    pairs: list[tuple[str, str]] = []
    for rp, lp in zip(rpaths, lpaths, strict=False):
        if Path(lp).is_dir():
            if not recursive:
                msg = f"Path is a directory: {rp}. Use --recursive to put directories."
                raise IsADirectoryError(msg)
            files = local_walk(lp, maxdepth=maxdepth)
            pairs.extend(rel_path_pairs(files, src_path=lp, dst_path=rp))
        else:
            pairs.append((lp, rp))
    return pairs


async def get_existing_url(urlpath: str) -> str:
    return urlpath


async def get_read_url(artifact: AsyncHyphaArtifact, params: dict[str, Any]) -> str:
    response = await artifact.get_client().get(
        get_method_url(artifact, ArtifactMethod.GET_FILE),
        params=params,
        headers=get_headers(artifact),
        timeout=60,
    )

    check_errors(response)

    return response.content.decode().strip('"')


async def get_write_url(artifact: AsyncHyphaArtifact, params: dict[str, Any]) -> str:
    response = await artifact.get_client().post(
        get_method_url(artifact, ArtifactMethod.PUT_FILE),
        json=params,
        headers=get_headers(artifact),
        timeout=60,
    )

    check_errors(response)

    return response.content.decode().strip('"')


async def walk_dir(
    self: AsyncHyphaArtifact,
    current_path: str,
    maxdepth: int | None,
    current_depth: int,
    version: str | None = None,
    *,
    withdirs: bool,
) -> dict[str, ArtifactItem]:
    """Recursively walk a directory."""
    results: dict[str, ArtifactItem] = {}

    try:
        items = await self.ls(current_path, version=version, detail=True)
    except (OSError, FileNotFoundError, httpx.RequestError):
        return {}

    for item in items:
        item_type = item["type"]
        item_name = item["name"]

        if item_type == "file" or (withdirs and item_type == "directory"):
            full_path = Path(current_path) / str(item_name)
            results[str(full_path)] = item

        if item_type == "directory" and (maxdepth is None or current_depth < maxdepth):
            subdir_path = Path(current_path) / str(item_name)
            subdirectory_results = await walk_dir(
                self,
                str(subdir_path),
                maxdepth,
                current_depth + 1,
                version=version,
                withdirs=withdirs,
            )
            results.update(subdirectory_results)

    return results


async def put_single_file(
    self: AsyncHyphaArtifact,
    src_path: str,
    dst_path: str,
) -> None:
    """Copy a single file from local to remote."""
    _lf = await aio_open(src_path, "rb")
    async with _lf as local_file:
        content = await local_file.read()

    async with self.open(dst_path, "wb") as remote_file:
        await remote_file.write(content)


def local_walk(
    src_path: str,
    maxdepth: int | None = None,
) -> list[str]:
    """Find all files in a local directory."""
    files: list[str] = []
    for root, _, dir_files in os.walk(src_path):
        if maxdepth is not None:
            rel_path = Path(root).relative_to(src_path)
            if len(rel_path.parts) >= maxdepth:
                continue
        files.extend(str(Path(root) / file_name) for file_name in dir_files)

    return files


def rel_path_pairs(
    files: list[str],
    src_path: str,
    dst_path: str,
) -> list[tuple[str, str]]:
    file_pairs: list[tuple[str, str]] = []
    for f in files:
        rel = Path(f).relative_to(src_path)
        file_pairs.append((f, str(dst_path / rel)))

    return file_pairs


def ensure_equal_len(
    rpath: str | list[str],
    lpath: str | list[str],
) -> tuple[list[str], list[str]]:
    """Assert that two paths (or lists of paths) are of equal length.

    Args:
        rpath (str | list[str]): The remote path(s) to check.
        lpath (str | list[str]): The local path(s) to check.

    Raises:
        ValueError: If the lengths of the paths do not match.
        ValueError: If the types of the paths do not match.

    Returns:
        _type_: _description_

    """
    if isinstance(rpath, str) and isinstance(lpath, str):
        rpath = [rpath]
        lpath = [lpath]
    elif isinstance(rpath, list) and isinstance(lpath, list):
        if len(rpath) != len(lpath):
            error_msg = "Both rpath and lpath must be the same length."
            raise ValueError(
                error_msg,
            )
    else:
        error_msg = "Both rpath and lpath must be strings or lists of strings."
        raise TypeError(
            error_msg,
        )

    return rpath, lpath


async def upload_part(
    self: AsyncHyphaArtifact,
    part_info: PreparedPartInfo,
) -> CompletedPart:
    """Upload a single part."""
    part_number = part_info["part_number"]
    upload_url = part_info["url"]

    async with self.open(upload_url, "wb") as f:
        await f.write(part_info["chunk"])

    etag = f.etag

    if etag is None:
        error_msg = "Failed to retrieve ETag from response"
        raise ValueError(error_msg)

    # Get ETag from response
    return CompletedPart(part_number=part_number, etag=etag)


def read_chunks(
    file_path: Path,
    chunk_size: int,
) -> list[bytes]:
    """Read file in chunks."""
    chunks: list[bytes] = []
    with file_path.open("rb") as f:
        while True:
            chunk_data = f.read(chunk_size)
            if not chunk_data:
                break
            chunks.append(chunk_data)

    return chunks


def should_use_multipart(
    local_path: Path,
    multipart_config: dict[str, Any] | None = None,
) -> bool:
    """Determine if multipart upload should be used."""
    file_size = local_path.stat().st_size

    if file_size > MAXIMUM_MULTIPART_THRESHOLD:
        return True

    if not multipart_config:
        return False

    chunk_size = multipart_config.get("chunk_size", DEFAULT_CHUNK_SIZE)

    if file_size < chunk_size:
        return False

    threshold = multipart_config.get("threshold")

    if threshold and file_size >= threshold:
        return True

    return bool(multipart_config.get("enable", False))


def validate_chunk_size(
    chunk_size: int,
) -> None:
    """Handle input errors for multipart upload.

    Args:
        file_size (int): The size of the local file in bytes.
        chunk_size (int): The chunk size for the upload.
        multipart_config (dict[str, Any]): The multipart configuration.

    Raises:
        ValueError: If the input parameters are invalid.

    """
    if chunk_size < MINIMUM_CHUNK_SIZE:
        error_msg = (
            "Chunk size must be greater than"
            f" {MINIMUM_CHUNK_SIZE // (1024 * 1024)}"
            "MB for multipart upload"
        )
        raise ValueError(error_msg)


async def start_multipart_upload(
    self: AsyncHyphaArtifact,
    local_path: Path,
    remote_path: str,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    download_weight: float = 1.0,
) -> dict[str, Any]:
    """Start a multipart upload for a file."""
    chunk_size = min(chunk_size, MAXIMUM_MULTIPART_THRESHOLD)
    file_size = local_path.stat().st_size
    validate_chunk_size(chunk_size)
    part_count = math.ceil(file_size / chunk_size)

    start_params = params_put_file_start_multipart(
        file_path=remote_path,
        part_count=part_count,
        download_weight=download_weight,
        use_proxy=self.use_proxy,
        use_local_url=self.use_local_url,
    )
    start_params = prepare_params(self, start_params)

    start_url = get_method_url(self, ArtifactMethod.PUT_FILE_START_MULTIPART)
    start_resp = await self.get_client().post(
        start_url,
        headers=get_headers(self),
        json=start_params,
    )
    check_errors(start_resp)
    return start_resp.json()


async def upload_with_callback(
    self: AsyncHyphaArtifact,
    semaphore: asyncio.Semaphore,
    pinfo: PreparedPartInfo,
    callback: Callable[[dict[str, Any]], None] | None,
    mpm: MultipartStatusMessage | None = None,
) -> CompletedPart:
    if callback and mpm:
        callback(mpm.part_info(pinfo["part_number"], pinfo.get("part_size")))
    try:
        async with semaphore:
            res = await upload_part(self, pinfo)
    except Exception as e:
        if callback and mpm:
            callback(mpm.part_error(pinfo["part_number"], str(e)))
        raise
    else:
        if callback and mpm:
            callback(mpm.part_success(pinfo["part_number"], pinfo.get("part_size")))
        return res


async def upload_parts(
    self: AsyncHyphaArtifact,
    local_path: Path,
    chunk_size: int,
    parts: list[UploadPartServerInfo],
    max_parallel_uploads: int,
    *,
    callback: Callable[[dict[str, Any]], None] | None = None,
    file_path: str | None = None,
) -> list[CompletedPart]:
    """Upload parts of a file in parallel.

    Args:
        self (AsyncHyphaArtifact): The artifact instance.
        local_path (Path): The local file path.
        chunk_size (int): The size of each chunk.
        parts (list[dict[str, Any]]): The list of parts to upload.
        max_parallel_uploads (int): Maximum number of concurrent part uploads.
        callback (Callable[[dict[str, Any]], None] | None): Optional progress callback
            invoked for each part with multipart status messages.
        file_path (str | None): Optional path used to annotate status messages.

    Returns:
        list[dict[str, Any]]: The list of responses from the uploaded parts.

    """
    chunks = read_chunks(local_path, chunk_size)
    enumerate_parts = enumerate(zip(parts, chunks, strict=False))
    parts_info: list[PreparedPartInfo] = [
        {
            "chunk": chunk,
            "url": part_info["url"],
            "part_number": part_info.get("part_number", index + 1),
            "part_size": len(chunk),
        }
        for index, (part_info, chunk) in enumerate_parts
    ]

    semaphore = asyncio.Semaphore(max_parallel_uploads)
    mpm = (
        MultipartStatusMessage("upload", file_path or str(local_path), len(parts_info))
        if callback is not None
        else None
    )

    upload_tasks = [
        upload_with_callback(self, semaphore, part_info, callback=callback, mpm=mpm)
        for part_info in parts_info
    ]

    return await asyncio.gather(*upload_tasks)


async def complete_multipart_upload(
    self: AsyncHyphaArtifact,
    upload_id: str,
    completed_parts: list[CompletedPart],
) -> None:
    """Complete a multipart upload.

    Args:
        self (AsyncHyphaArtifact): The artifact instance.
        upload_id (str): The ID of the upload.
        completed_parts (list[dict[str, Any]]): The list of completed parts.

    """
    simple_params = params_put_file_complete_multipart(
        upload_id=upload_id,
        parts=completed_parts,
    )
    complete_params = prepare_params(self, simple_params)
    complete_url = get_method_url(self, ArtifactMethod.PUT_FILE_COMPLETE_MULTIPART)
    complete_resp = await self.get_client().post(
        complete_url,
        json=complete_params,
        headers=get_headers(self),
    )
    check_errors(complete_resp)


def get_multipart_settings(
    multipart_config: dict[str, Any] | None = None,
) -> tuple[int, int]:
    """Get the default multipart settings."""
    if multipart_config is None:
        return DEFAULT_CHUNK_SIZE, 4

    chunk_size = multipart_config.get("chunk_size", DEFAULT_CHUNK_SIZE)
    max_parallel_uploads = multipart_config.get("max_parallel_uploads", 4)

    return chunk_size, max_parallel_uploads


async def upload_multipart(
    self: AsyncHyphaArtifact,
    local_path: Path,
    remote_path: str,
    chunk_size: int = DEFAULT_CHUNK_SIZE,
    max_parallel_uploads: int = 4,
    download_weight: float = 1.0,
    *,
    callback: Callable[[dict[str, Any]], None] | None = None,
) -> None:
    """Upload a file using multipart upload with parallel uploads."""
    multipart_info = await start_multipart_upload(
        self,
        local_path,
        remote_path,
        chunk_size=chunk_size,
        download_weight=download_weight,
    )

    parts = multipart_info["parts"]
    completed_parts = await upload_parts(
        self,
        local_path,
        chunk_size,
        parts,
        max_parallel_uploads,
        callback=callback,
        file_path=str(local_path),
    )

    upload_id = multipart_info["upload_id"]
    await complete_multipart_upload(self, upload_id, completed_parts)


def prepare_params(
    self: AsyncHyphaArtifact,
    params: Mapping[str, object] | None = None,
) -> dict[str, Any]:
    """Extend parameters with artifact_id."""
    cleaned_params: dict[str, object] = {
        k: v for k, v in (dict(params or {})).items() if v is not None
    }
    cleaned_params["artifact_id"] = self.artifact_id
    return cleaned_params


def get_method_url(self: AsyncHyphaArtifact, method: ArtifactMethod) -> str:
    """Get the URL for a specific artifact method."""
    return f"{self.artifact_url}/{method}"


def get_headers(self: AsyncHyphaArtifact) -> dict[str, str]:
    """Get headers for HTTP requests.

    Returns:
        dict[str, str]: Headers to include in the request.

    """
    return {"Authorization": f"Bearer {self.token}"} if self.token else {}


def check_errors(response: httpx.Response) -> None:
    """Handle errors in HTTP responses."""
    if response.status_code != HTTPStatus.OK:
        error_msg = f"Unexpected error: {response.text}"
        raise httpx.RequestError(error_msg)

    response.raise_for_status()



"""Serves as a backward-compatibility layer for the AsyncHyphaArtifact class.

The AsyncHyphaArtifact class has been refactored into a module.
This file imports the class from the new module to ensure that existing code
that imports from this file continues to work.
"""

from .async_hypha_artifact import AsyncHyphaArtifact

__all__ = ["AsyncHyphaArtifact"]



"""Integration tests for artifact state methods and versioned retrievals.

Covers:
- Async create() and delete() called without parameters.
- Versioned reads across v0 and a newer version for common read methods.
"""

from __future__ import annotations

import asyncio
import uuid

import pytest
import pytest_asyncio

from hypha_artifact import AsyncHyphaArtifact

# pylint: disable=protected-access,redefined-outer-name,broad-except


@pytest_asyncio.fixture
async def ephemeral_artifact(credentials: tuple[str, str]):
    """Yield a brand-new AsyncHyphaArtifact with a unique alias for isolated tests."""
    token, workspace = credentials
    alias = f"test-state-{uuid.uuid4().hex[:8]}"
    artifact = AsyncHyphaArtifact(
        alias,
        workspace=workspace,
        token=token,
        server_url="https://hypha.aicell.io",
    )
    try:
        yield artifact
    finally:
        await artifact.aclose()


class TestAsyncStateMethods:
    """Integration tests for Async create() and delete() methods."""

    @pytest.mark.asyncio
    async def test_create_without_params(self, ephemeral_artifact: AsyncHyphaArtifact):
        """Calling create() with no parameters should succeed and allow listing root."""
        await ephemeral_artifact.create()

        # Basic smoke: can list root on a newly created artifact
        files = await ephemeral_artifact.ls("/", detail=True)
        assert isinstance(files, list)

        # Cleanup for this test
        await ephemeral_artifact.delete()

    @pytest.mark.asyncio
    async def test_delete_without_params(self, ephemeral_artifact: AsyncHyphaArtifact):
        """Delete with no parameters should remove the entire artifact."""
        # Create first so we can delete it
        await ephemeral_artifact.create()

        # Deleting the entire artifact (default behavior)
        await ephemeral_artifact.delete()

        # Subsequent operations against the deleted artifact should fail
        with pytest.raises(Exception):
            await ephemeral_artifact.ls("/", detail=True)


class TestVersionedRetrievals:
    """Integration tests that verify the version parameter on read methods."""

    @pytest.mark.asyncio
    async def test_version_parameter_across_methods(self, credentials: tuple[str, str]):
        token, workspace = credentials
        alias = f"test-versions-{uuid.uuid4().hex[:8]}"
        artifact = AsyncHyphaArtifact(
            alias,
            workspace=workspace,
            token=token,
            server_url="https://hypha.aicell.io",
        )

        try:
            # 1) Create artifact -> should create v0 (metadata only)
            await artifact.create()

            # 2) Add a file to v0 (stage and commit without version intent -> updates latest v0)
            fname = "verfile.txt"
            content_v0 = "A-version"
            await artifact.edit(stage=True)
            async with artifact.open(fname, "w") as f:
                await f.write(content_v0)
            await artifact.commit(comment="seed v0 contents")

            # Sanity checks on v0
            assert await artifact.exists(fname, version="v0") is True
            cat_v0 = await artifact.cat(fname, version="v0")
            assert cat_v0 == content_v0

            # 3) Create a new version and change the file content
            content_v1 = "B-version"
            await artifact.edit(stage=True, version="new")
            async with artifact.open(fname, "w") as f:
                await f.write(content_v1)
            await artifact.commit(comment="create new version with updated content")

            # Latest should return v1 content; explicit v0 should return old content
            latest_cat = await artifact.cat(fname)
            assert latest_cat == content_v1
            explicit_v0_cat = await artifact.cat(fname, version="v0")
            assert explicit_v0_cat == content_v0

            # ls with version should see the file in both versions
            names_latest = [i["name"] for i in await artifact.ls("/", detail=True)]
            assert fname in names_latest
            names_v0 = [
                i["name"] for i in await artifact.ls("/", detail=True, version="v0")
            ]
            assert fname in names_v0

            # info/size consistency across versions
            info_latest = await artifact.info(fname)
            info_v0 = await artifact.info(fname, version="v0")
            assert info_latest.get("size") == len(content_v1)
            assert info_v0.get("size") == len(content_v0)

            # head should reflect per-version content
            head_latest = await artifact.head(fname, size=2)
            head_v0 = await artifact.head(fname, size=2, version="v0")
            assert head_latest == content_v1[:2].encode()
            assert head_v0 == content_v0[:2].encode()

        finally:
            # Cleanup: remove the whole artifact
            try:
                await artifact.delete()
            except Exception:
                pass
            await artifact.aclose()


class TestListChildren:
    """Integration tests for listing child artifacts within a collection."""

    @pytest.mark.asyncio
    async def test_list_children_basic_and_ordering(
        self,
        credentials: tuple[str, str],
    ) -> None:
        token, workspace = credentials

        # Parent collection
        coll_alias = f"test-coll-{uuid.uuid4().hex[:8]}"
        coll = AsyncHyphaArtifact(
            coll_alias,
            workspace=workspace,
            token=token,
            server_url="https://hypha.aicell.io",
        )

        # Two committed children under the collection
        child1_alias = f"alpha-{uuid.uuid4().hex[:4]}"
        child2_alias = f"beta-{uuid.uuid4().hex[:4]}"

        child1 = AsyncHyphaArtifact(
            child1_alias,
            workspace=workspace,
            token=token,
            server_url="https://hypha.aicell.io",
        )
        child2 = AsyncHyphaArtifact(
            child2_alias,
            workspace=workspace,
            token=token,
            server_url="https://hypha.aicell.io",
        )

        try:
            # Create the parent collection (committed v0)
            await coll.create(
                type="collection",
                manifest={"name": coll_alias, "collection": []},
            )

            parent_id = f"{workspace}/{coll_alias}"

            # Create two committed children (default create -> committed v0)
            await child1.create(
                parent_id=parent_id,
                manifest={"name": "Alpha", "likes": 10},
            )
            await child2.create(
                parent_id=parent_id,
                manifest={"name": "Beta", "likes": 5},
            )

            # Basic listing
            # Retry a few times in case of eventual consistency
            async def _list_children_committed():
                return await coll.list_children(stage=False)

            for _ in range(5):
                res = await _list_children_committed()
                # Normalize and break early if both names present
                cand = (res.get("items") if isinstance(res, dict) else res) or []
                cand_names = {
                    i.get("manifest", {}).get("name")
                    for i in cand
                    if isinstance(i, dict)
                }
                if {"Alpha", "Beta"}.issubset(cand_names):
                    break
                await asyncio.sleep(0.3)

            res = await _list_children_committed()

            # Result may be a list or a dict with items/results; normalize
            if isinstance(res, dict):
                items = (
                    res.get("items")
                    or res.get("results")
                    or res.get("data")
                    or res.get("artifacts")
                    or []
                )
            else:
                items = res

            assert isinstance(items, list)
            names = {
                i.get("manifest", {}).get("name") for i in items if isinstance(i, dict)
            }
            assert {"Alpha", "Beta"}.issubset(names)

            # Ordering by custom JSON field (descending by likes)
            res_ordered = await coll.list_children(
                order_by="manifest.likes>",
                stage=False,
            )
            if isinstance(res_ordered, dict):
                ordered_items = (
                    res_ordered.get("items")
                    or res_ordered.get("results")
                    or res_ordered.get("data")
                    or res_ordered.get("artifacts")
                    or []
                )
            else:
                ordered_items = res_ordered

            # Ensure we have at least the two children and ordering is respected
            ordered_names = [
                i.get("manifest", {}).get("name")
                for i in ordered_items
                if isinstance(i, dict)
            ]
            # Since likes: Alpha=10, Beta=5 and '>' means descending, Alpha should come before Beta
            if set(["Alpha", "Beta"]).issubset(set(ordered_names)):
                alpha_idx = ordered_names.index("Alpha")
                beta_idx = ordered_names.index("Beta")
                assert alpha_idx < beta_idx

        finally:
            # Cleanup: delete parent recursively (children included)
            try:
                await coll.delete(delete_files=True, recursive=True)
            except Exception:
                pass
            await coll.aclose()
            await child1.aclose()
            await child2.aclose()

    @pytest.mark.asyncio
    async def test_list_children_with_keywords_and_filters_and_stage(
        self,
        credentials: tuple[str, str],
    ) -> None:
        token, workspace = credentials

        coll_alias = f"test-coll-{uuid.uuid4().hex[:8]}"
        staged_child_alias = f"staged-{uuid.uuid4().hex[:4]}"
        committed_child_alias = f"commit-{uuid.uuid4().hex[:4]}"

        coll = AsyncHyphaArtifact(
            coll_alias,
            workspace=workspace,
            token=token,
            server_url="https://hypha.aicell.io",
        )
        staged_child = AsyncHyphaArtifact(
            staged_child_alias,
            workspace=workspace,
            token=token,
            server_url="https://hypha.aicell.io",
        )
        committed_child = AsyncHyphaArtifact(
            committed_child_alias,
            workspace=workspace,
            token=token,
            server_url="https://hypha.aicell.io",
        )

        try:
            await coll.create(
                type="collection",
                manifest={"name": coll_alias, "collection": []},
            )
            parent_id = f"{workspace}/{coll_alias}"

            # One committed child
            await committed_child.create(
                parent_id=parent_id,
                manifest={"name": "Gamma", "category": "x"},
            )

            # One staged child (do not commit)
            await staged_child.create(
                parent_id=parent_id,
                manifest={"name": "Delta", "category": "y"},
                version="stage",
            )

            # Keywords should match by name
            # Committed-only listing should find Gamma
            kw_res = await coll.list_children(keywords=["Gamma"], stage=False)
            kw_items = kw_res.get("items") if isinstance(kw_res, dict) else kw_res
            assert kw_items
            kw_names = {
                i.get("manifest", {}).get("name")
                for i in kw_items
                if isinstance(i, dict)
            }
            assert "Gamma" in kw_names

            # Filters against manifest fields
            flt_res = await coll.list_children(
                filters={"manifest": {"category": "x"}},
                stage=False,
            )
            flt_items = flt_res.get("items") if isinstance(flt_res, dict) else flt_res
            assert flt_items
            flt_names = {
                i.get("manifest", {}).get("name")
                for i in flt_items
                if isinstance(i, dict)
            }
            assert "Gamma" in flt_names
            assert (
                "Delta" not in flt_names
            )  # staged child shouldn't appear without stage=True

            # Stage-only listing should include the staged child
            stage_only = await coll.list_children(stage=True)
            s_items = (
                stage_only.get("items") if isinstance(stage_only, dict) else stage_only
            )
            assert s_items
            s_names = {
                i.get("manifest", {}).get("name")
                for i in s_items
                if isinstance(i, dict)
            }
            assert "Delta" in s_names
            # And the committed one should not be present when stage=True
            assert "Gamma" not in s_names

        finally:
            try:
                await coll.delete(delete_files=True, recursive=True)
            except Exception:
                pass
            await coll.aclose()
            await staged_child.aclose()
            await committed_child.aclose()
